@@ -40,7 +40,7 @@ def default(val, d):
 def divisible_by(numer, denom):
     return (numer % denom) == 0
 
-def safe_div(numer, denom, eps = 1e-8):
+def safe_div(numer, denom, eps = 1e-10):
     return numer / denom.clamp(min = eps)
 
 def cycle(dl):
@@ -64,6 +64,9 @@ def convert_image_to(img_type, image):
         return image.convert(img_type)
     return image
 
+def Sequential(*mods):
+    return nn.Sequential(*filter(exists, mods))
+
 # use layernorm without bias, more stable
 
 class LayerNorm(nn.Module):
@@ -346,7 +349,8 @@ class RIN(nn.Module):
         dim_latent = None,              # will default to image dim (dim)
         num_latents = 256,              # they still had to use a fair amount of latents for good results (256), in line with the Perceiver line of papers from Deepmind
         learned_sinusoidal_dim = 16,
-        latent_token_time_cond = True,  # whether to use 1 latent token as time conditioning, or do it the adaptive layernorm way (which is highly effective as shown by some other papers "Paella" - Dominic Rampas et al.)
+        latent_token_time_cond = False, # whether to use 1 latent token as time conditioning, or do it the adaptive layernorm way (which is highly effective as shown by some other papers "Paella" - Dominic Rampas et al.)
+        dual_patchnorm = True,
         **attn_kwargs
     ):
         super().__init__()
@@ -378,12 +382,36 @@ class RIN(nn.Module):
 
         # pixels to patch and back
 
-        self.to_patches = nn.Sequential(
+        self.to_patches = Sequential(
             Rearrange('b c (h p1) (w p2) -> b (h w) (c p1 p2)', p1 = patch_size, p2 = patch_size),
-            nn.Linear(pixel_patch_dim * 2, dim)
+            nn.LayerNorm(pixel_patch_dim * 2) if dual_patchnorm else None,
+            nn.Linear(pixel_patch_dim * 2, dim),
+            nn.LayerNorm(dim) if dual_patchnorm else None,
         )
 
-        self.axial_pos_emb = nn.Parameter(torch.randn(2, patch_height_width, dim) * 0.02)
+        # axial positional embeddings, parameterized by an MLP
+
+        pos_emb_dim = dim // 2
+
+        self.axial_pos_emb_height_mlp = nn.Sequential(
+            Rearrange('... -> ... 1'),
+            nn.Linear(1, pos_emb_dim),
+            nn.SiLU(),
+            nn.Linear(pos_emb_dim, pos_emb_dim),
+            nn.SiLU(),
+            nn.Linear(pos_emb_dim, dim)
+        )
+
+        self.axial_pos_emb_width_mlp = nn.Sequential(
+            Rearrange('... -> ... 1'),
+            nn.Linear(1, pos_emb_dim),
+            nn.SiLU(),
+            nn.Linear(pos_emb_dim, pos_emb_dim),
+            nn.SiLU(),
+            nn.Linear(pos_emb_dim, dim)
+        )
+
+        # nn.Parameter(torch.randn(2, patch_height_width, dim) * 0.02)
 
         self.to_pixels = nn.Sequential(
             LayerNorm(dim),
@@ -408,6 +436,10 @@ class RIN(nn.Module):
 
         self.blocks = nn.ModuleList([RINBlock(dim, dim_latent = dim_latent, latent_self_attn_depth = latent_self_attn_depth, **attn_kwargs) for _ in range(depth)])
 
+    @property
+    def device(self):
+        return next(self.parameters()).device
+
     def forward(
         self,
         x,
@@ -445,7 +477,9 @@ class RIN(nn.Module):
 
         patches = self.to_patches(x)
 
-        pos_emb_h, pos_emb_w = self.axial_pos_emb
+        height_range = width_range = torch.linspace(0., 1., steps = int(math.sqrt(patches.shape[-2])), device = self.device)
+        pos_emb_h, pos_emb_w = self.axial_pos_emb_height_mlp(height_range), self.axial_pos_emb_width_mlp(width_range)
+
         pos_emb = rearrange(pos_emb_h, 'i d -> i 1 d') + rearrange(pos_emb_w, 'j d -> 1 j d')
         patches = patches + rearrange(pos_emb, 'i j d -> (i j) d')
 
@@ -464,7 +498,7 @@ class RIN(nn.Module):
         # remove time conditioning token, if that is the settings
 
         if self.latent_token_time_cond:
-            latents = latents[:, 1:]
+            latents = latents[:, :-1]
 
         return pixels, latents
 
@@ -531,9 +565,11 @@ class GaussianDiffusion(nn.Module):
         timesteps = 1000,
         use_ddim = True,
         noise_schedule = 'sigmoid',
-        objective = 'eps',
+        objective = 'v',
         schedule_kwargs: dict = dict(),
         time_difference = 0.,
+        min_snr_loss_weight = True,
+        min_snr_gamma = 5,
         train_prob_self_cond = 0.9,
         scale = 1.                      # this will be set to < 1. for better convergence when training on higher resolution images
     ):
@@ -541,7 +577,7 @@ class GaussianDiffusion(nn.Module):
         self.model = model
         self.channels = self.model.channels
 
-        assert objective in {'x0', 'eps'}, 'objective must be either predict x0 or noise'
+        assert objective in {'x0', 'eps', 'v'}, 'objective must be either predict x0 or noise'
         self.objective = objective
 
         self.image_size = model.image_size
@@ -577,6 +613,11 @@ class GaussianDiffusion(nn.Module):
 
         self.train_prob_self_cond = train_prob_self_cond
 
+        # min snr loss weight
+
+        self.min_snr_loss_weight = min_snr_loss_weight
+        self.min_snr_gamma = min_snr_gamma
+
     @property
     def device(self):
         return next(self.model.parameters()).device
@@ -633,6 +674,9 @@ class GaussianDiffusion(nn.Module):
             elif self.objective == 'eps':
                 x_start = safe_div(img - sigma * model_output, alpha)
 
+            elif self.objective == 'v':
+                x_start = alpha * img - sigma * model_output
+
             # clip x0
 
             x_start.clamp_(-1., 1.)
@@ -701,17 +745,16 @@ class GaussianDiffusion(nn.Module):
             elif self.objective == 'eps':
                 x_start = safe_div(img - sigma * model_output, alpha)
 
+            elif self.objective == 'v':
+                x_start = alpha * img - sigma * model_output
+
             # clip x0
 
             x_start.clamp_(-1., 1.)
 
             # get predicted noise
 
-            if self.objective == 'x0':
-                pred_noise = safe_div(img - alpha * x_start, sigma)
-
-            elif self.objective == 'eps':
-                pred_noise = model_output
+            pred_noise = safe_div(img - alpha * x_start, sigma)
 
             # calculate x next
 
@@ -765,6 +808,10 @@ class GaussianDiffusion(nn.Module):
                 elif self.objective == 'eps':
                     self_cond = safe_div(noised_img - sigma * model_output, alpha)
 
+                elif self.objective == 'v':
+                    self_cond = alpha * noised_img - sigma * model_output
+
+                self_cond.clamp_(-1., 1.)
                 self_cond = self_cond.detach()
 
         # predict and take gradient step
@@ -777,7 +824,30 @@ class GaussianDiffusion(nn.Module):
         elif self.objective == 'eps':
             target = noise
 
-        return F.mse_loss(pred, target)
+        elif self.objective == 'v':
+            target = alpha * noise - sigma * img
+
+        loss = F.mse_loss(pred, target, reduction = 'none')
+        loss = reduce(loss, 'b ... -> b', 'mean')
+
+        # min snr loss weight
+
+        snr = (alpha * alpha) / (sigma * sigma)
+        maybe_clipped_snr = snr.clone()
+
+        if self.min_snr_loss_weight:
+            maybe_clipped_snr.clamp_(min = self.min_snr_gamma)
+
+        if self.objective == 'x0':
+            loss_weight = maybe_clipped_snr / snr
+
+        elif self.objective == 'eps':
+            loss_weight = maybe_clipped_snr
+
+        elif self.objective == 'v':
+            loss_weight = maybe_clipped_snr / (snr + 1)
+
+        return (loss * loss_weight).mean()
 
 # dataset classes
 
@@ -829,7 +899,7 @@ class Trainer(object):
         train_num_steps = 100000,
         ema_update_every = 10,
         ema_decay = 0.995,
-        adam_betas = (0.9, 0.99),
+        betas = (0.9, 0.99),
         save_and_sample_every = 1000,
         num_samples = 25,
         results_folder = './results',
@@ -869,7 +939,7 @@ class Trainer(object):
 
         # optimizer
 
-        self.opt = Adam(diffusion_model.parameters(), lr = train_lr, betas = adam_betas)
+        self.opt = Adam(diffusion_model.parameters(), lr = train_lr, betas = betas)
 
         # for logging results in a folder periodically
 
